from __future__ import annotations

import enum
import math

import zigpy.types


class PackedUInt21(zigpy.types.uint_t, bits=21):  # type: ignore[call-arg]
    def serialize(self) -> bytes:
        n = int(self)
        chunks = []

        while n:
            # Set the least significant bit on all other octets.
            chunks.append((n & 0b01111111) | 0b10000000)
            n >>= 7

        # Clear the most significant bit of the most significant octet.
        chunks[-1] &= 0b01111111

        return bytes(chunks)

    @classmethod
    def deserialize(cls, data: bytes) -> tuple[PackedUInt21, bytes]:
        chunks = []

        for byte in data:
            chunks.append(byte & 0b01111111)

            if len(chunks) > math.ceil(cls._bits / 8):
                raise ValueError(
                    f"Packed integer cannot be larger than {cls.max_value}"
                )

            if byte & 0b10000000 == 0:
                break

        n = 0

        for chunk in chunks[::-1]:
            n = (n << 7) | chunk

        return cls(n), data[len(chunks) :]


class CommandID(PackedUInt21, enum.Enum):
    NOOP = 0
    RESET = 1
    PROP_VALUE_GET = 2
    PROP_VALUE_SET = 3
    PROP_VALUE_INSERT = 4
    PROP_VALUE_REMOVE = 5
    PROP_VALUE_IS = 6
    PROP_VALUE_INSERTED = 7
    PROP_VALUE_REMOVED = 8

    PEEK = 18
    PEEK_RET = 19
    POKE = 20
    PROP_VALUE_MULTI_GET = 21
    PROP_VALUE_MULTI_SET = 22
    PROP_VALUES_ARE = 23

    NET_SAVE = 9
    NET_CLEAR = 10
    NET_RECALL = 11

    HBO_OFFLOAD = 12
    HBO_RECLAIM = 13
    HBO_DROP = 14
    HBO_OFFLOADED = 15
    HBO_RECLAIMED = 16
    HBO_DROPPED = 17


class PropertyID(PackedUInt21, enum.Enum):
    # Core Properties
    LAST_STATUS = 0
    PROTOCOL_VERSION = 1
    NCP_VERSION = 2
    INTERFACE_TYPE = 3
    INTERFACE_VENDOR_ID = 4
    CAPS = 5
    INTERFACE_COUNT = 6
    POWER_STATE = 7
    HWADDR = 8
    LOCK = 9

    # Host Buffer Offload
    HBO_MEM_MAX = 10
    HBO_BLOCK_MAX = 11

    # Stream Properties
    STREAM_DEBUG = 112
    STREAM_RAW = 113
    STREAM_NET = 114
    # STREAM_NET_INSECURE = 114  # has the same ID as `STREAM_NET`?

    # PHY Properties
    PHY_ENABLED = 32
    PHY_CHAN = 33
    PHY_CHAN_SUPPORTED = 34
    PHY_FREQ = 35
    PHY_CCA_THRESHOLD = 36
    PHY_TX_POWER = 37
    PHY_RSSI = 38
    PHY_RX_SENSITIVITY = 39

    # MAC Properties
    MAC_SCAN_STATE = 38
    MAC_SCAN_MASK = 49
    MAC_SCAN_PERIOD = 50
    MAC_SCAN_BEACON = 51
    MAC_15_4_LADDR = 52
    MAC_15_4_SADDR = 53
    MAC_15_4_PANID = 54
    MAC_RAW_STREAM_ENABLED = 55
    MAC_PROMISCUOUS_MODE = 56
    MAC_ENERGY_SCAN_RESULT = 57
    MAC_WHITELIST = 4864
    MAC_WHITELIST_ENABLED = 4865

    # NET Properties
    NET_SAVED = 64
    NET_IF_UP = 65
    NET_STACK_UP = 66
    NET_ROLE = 67
    NET_NETWORK_NAME = 68
    NET_XPANID = 69
    NET_MASTER_KEY = 70
    NET_KEY_SEQUENCE_COUNTER = 71
    NET_PARTITION_ID = 72
    NET_REQUIRE_JOIN_EXISTING = 73
    NET_KEY_SWITCH_GUARDTIME = 74
    NET_PSKC = 75

    # IPv6 Properties
    IPV6_LL_ADDR = 96
    IPV6_ML_ADDR = 97
    IPV6_ML_PREFIX = 98
    IPV6_ADDRESS_TABLE = 99
    IPV6_ICMP_PING_OFFLOAD = 101

    # Debug Properties
    DEBUG_TEST_ASSERT = 16384
    DEBUG_NCP_LOG_LEVEL = 16385

    # Thread Properties
    THREAD_LEADER_ADDR = 80
    THREAD_PARENT = 81
    THREAD_CHILD_TABLE = 82
    THREAD_LEADER_RID = 83
    THREAD_LEADER_WEIGHT = 84
    THREAD_LOCAL_LEADER_WEIGHT = 85
    THREAD_NETWORK_DATA = 86
    THREAD_NETWORK_DATA_VERSION = 87
    THREAD_STABLE_NETWORK_DATA = 88
    THREAD_STABLE_NETWORK_DATA_VERSION = 89
    THREAD_ON_MESH_NETS = 90
    THREAD_LOCAL_ROUTES = 91
    THREAD_ASSISTING_PORTS = 92
    THREAD_ALLOW_LOCAL_NET_DATA_CHANGE = 93
    THREAD_MODE = 94
    THREAD_CHILD_TIMEOUT = 5376
    THREAD_RLOC16 = 5377
    THREAD_ROUTER_UPGRADE_THRESHOLD = 5378
    THREAD_CONTEXT_REUSE_DELAY = 5379
    THREAD_NETWORK_ID_TIMEOUT = 5380
    THREAD_ACTIVE_ROUTER_IDS = 5381
    THREAD_RLOC16_DEBUG_PASSTHRU = 5382
    THREAD_ROUTER_ROLE_ENABLED = 5383
    THREAD_ROUTER_DOWNGRADE_THRESHOLD = 5384
    THREAD_ROUTER_SELECTION_JITTER = 5385
    THREAD_PREFERRED_ROUTER_ID = 5386
    THREAD_NEIGHBOR_TABLE = 5387
    THREAD_CHILD_COUNT_MAX = 5388
    THREAD_LEADER_NETWORK_DATA = 5389
    THREAD_STABLE_LEADER_NETWORK_DATA = 5390
    THREAD_JOINERS = 5391
    THREAD_COMMISSIONER_ENABLED = 5392
    THREAD_BA_PROXY_ENABLED = 5393
    THREAD_BA_PROXY_STREAM = 5394
    THREAD_DISOVERY_SCAN_JOINER_FLAG = 5395
    THREAD_DISCOVERY_SCAN_ENABLE_FILTERING = 5396
    THREAD_DISCOVERY_SCAN_PANID = 5397
    THREAD_STEERING_DATA = 5398

    # Jam detection
    JAM_DETECT_ENABLE = 4608
    JAM_DETECTED = 4609
    JAM_DETECT_RSSI_THRESHOLD = 4610
    JAM_DETECT_WINDOW = 4611
    JAM_DETECT_BUSY = 4612
    JAM_DETECT_HISTORY_BITMAP = 4613

    # GPIO
    GPIO_CONFIG = 4096
    GPIO_STATE = 4098
    GPIO_STATE_SET = 4099
    GPIO_STATE_CLEAR = 4100

    # True random number generation
    TRNG_32 = 4101
    TRNG_128 = 4102
    TRNG_RAW_32 = 4103


class ResetReason(zigpy.types.enum8):
    PLATFORM = 1
    STACK = 2
    BOOTLOADER = 3


class HDLCSpecial(enum.IntEnum):
    FLAG = 0x7E
    ESCAPE = 0x7D
    XON = 0x11
    XOFF = 0x13
    VENDOR = 0xF8



import enum


class FirmwareImageType(enum.Enum):
    ZIGBEE_NCP = "zigbee_ncp"
    OPENTHREAD_RCP = "openthread_rcp"
    ZWAVE_NCP = "zwave_ncp"
    BOOTLOADER = "bootloader"
    MULTIPAN = "multipan"

    UNKNOWN = "unknown"


LEGACY_FIRMWARE_TYPE_REMAPPING = {
    "ncp-uart-hw": FirmwareImageType.ZIGBEE_NCP,
    "ncp-uart-sw": FirmwareImageType.ZIGBEE_NCP,
    "rcp-uart-802154": FirmwareImageType.MULTIPAN,
    "ot-rcp": FirmwareImageType.OPENTHREAD_RCP,
    "z-wave": FirmwareImageType.ZWAVE_NCP,
    "gecko-bootloader": FirmwareImageType.BOOTLOADER,
}


class ApplicationType(enum.Enum):
    GECKO_BOOTLOADER = "bootloader"
    CPC = "cpc"
    EZSP = "ezsp"
    SPINEL = "spinel"


FW_IMAGE_TYPE_TO_APPLICATION_TYPE = {
    FirmwareImageType.ZIGBEE_NCP: ApplicationType.EZSP,
    FirmwareImageType.MULTIPAN: ApplicationType.CPC,
    FirmwareImageType.OPENTHREAD_RCP: ApplicationType.SPINEL,
    FirmwareImageType.BOOTLOADER: ApplicationType.GECKO_BOOTLOADER,
}


DEFAULT_BAUDRATES = {
    ApplicationType.GECKO_BOOTLOADER: [115200],
    ApplicationType.CPC: [460800, 115200, 230400],
    ApplicationType.EZSP: [115200],
    ApplicationType.SPINEL: [460800],
}


class ResetTarget(enum.Enum):
    YELLOW = "yellow"
    IHOST = "ihost"
    SLZB07 = "slzb07"
    SONOFF = "sonoff"


GPIO_CONFIGS = {
    ResetTarget.YELLOW: {
        "chip": "/dev/gpiochip0",
        "pin_states": {
            24: [True, False, False, True],
            25: [True, False, True, True],
        },
        "toggle_delay": 0.1,
    },
    ResetTarget.IHOST: {
        "chip": "/dev/gpiochip1",
        "pin_states": {
            27: [True, False, False, True],
            26: [True, False, True, True],
        },
        "toggle_delay": 0.1,
    },
    ResetTarget.SLZB07: {
        "chip_name": "cp210x",
        "pin_states": {
            5: [True, False, False, True],
            4: [True, False, True, True],
        },
        "toggle_delay": 0.1,
    },
}



from __future__ import annotations

import asyncio
import dataclasses
import logging
import typing

import zigpy.types

from .common import asyncio_timeout, crc16_ccitt

_LOGGER = logging.getLogger(__name__)

BLOCK_SIZE = 128
RECEIVE_TIMEOUT = 2

_WRITER_GRAVEYARD: list[asyncio.StreamWriter] = []


class PacketType(zigpy.types.enum8):
    """XModem packet type byte."""

    SOH = 0x01  # Start of Header
    EOT = 0x04  # End of Transmission
    CAN = 0x18  # Cancel
    ETB = 0x17  # End of Transmission Block
    ACK = 0x06  # Acknowledge
    NAK = 0x15  # Not Acknowledge


@dataclasses.dataclass(frozen=True)
class XmodemCRCPacket:
    """XModem CRC packet implementing the zigpy `serialize` API."""

    number: zigpy.types.uint8_t
    payload: bytes

    def serialize(self) -> bytes:
        """Serialize the packet, computing header and payload checksums."""
        assert len(self.payload) == BLOCK_SIZE
        return (
            bytes([PacketType.SOH, self.number, 0xFF - self.number])
            + self.payload
            + crc16_ccitt(self.payload).to_bytes(2, "big")
        )


class ReceiverCancelled(Exception):
    """Receiver cancelled the transmission with a `CAN` status."""


async def send_xmodem128_crc_data(
    data: bytes,
    *,
    reader: asyncio.StreamReader,
    writer: asyncio.StreamReader,
    max_failures: int,
) -> None:
    """Send `data` to an XModem receiver until an ACK is received."""

    for attempt in range(max_failures + 1):
        # Send off the data
        _LOGGER.debug("Sending data %r (attempt %d)", data, attempt)
        writer.write(data)
        await writer.drain()

        # And wait for a response
        async with asyncio_timeout(RECEIVE_TIMEOUT):
            rsp_byte = await reader.readexactly(1)

        _LOGGER.debug("Got response: %r", rsp_byte)

        if rsp_byte[0] == PacketType.ACK:
            return
        elif rsp_byte[0] == PacketType.NAK:
            _LOGGER.debug("Got a NAK, retrying")

            if attempt >= max_failures:
                raise ValueError(f"Received {max_failures} consecutive failures")
        elif rsp_byte[0] == PacketType.CAN:
            raise ReceiverCancelled()
        else:
            raise ValueError(f"Invalid response: {rsp_byte!r}")


async def send_xmodem128_crc(
    data: bytes,
    *,
    transport: asyncio.Transport,
    max_failures: int = 3,
    progress_callback: typing.Callable[[int, int], typing.Any] | None = None,
) -> None:
    """Send `data` over `transport` using XModemCRC with a 128 byte block size."""

    if len(data) % BLOCK_SIZE != 0:
        raise ValueError(f"Data length must be divisible by {BLOCK_SIZE}: {len(data)}")

    loop = asyncio.get_running_loop()

    # Initialization of the reader and writer objects is from `asyncio.open_connection`
    reader = asyncio.StreamReader(limit=65536, loop=loop)
    protocol = asyncio.StreamReaderProtocol(reader, loop=loop)
    writer = asyncio.StreamWriter(transport, protocol, reader, loop)

    # Swap protocols
    old_protocol = transport.get_protocol()
    transport.set_protocol(protocol)

    try:
        # Read until the first ASCII "C"
        await reader.readuntil(b"C")

        if progress_callback is not None:
            progress_callback(0, len(data))

        # FIXME: ensure any subsequent "C"s have been cleared so they do not interfere
        reader._buffer.clear()

        for index in range(0, len(data) // BLOCK_SIZE):
            packet = XmodemCRCPacket(
                number=(index + 1) & 0xFF,  # `seq` starts at 1 and then wraps
                payload=data[BLOCK_SIZE * index : BLOCK_SIZE * (index + 1)],
            )

            # Send the packet
            await send_xmodem128_crc_data(
                data=packet.serialize(),
                reader=reader,
                writer=writer,
                max_failures=max_failures,
            )

            offset = (index + 1) * BLOCK_SIZE

            if progress_callback is not None:
                progress_callback(offset, len(data))

            _LOGGER.debug("Firmware upload progress: %0.2f%%", 100 * offset / len(data))

        # Once we are done, finalize the upload
        await send_xmodem128_crc_data(
            data=bytes([PacketType.EOT]),
            reader=reader,
            writer=writer,
            max_failures=max_failures,
        )
    finally:
        # XXX: Make sure the writer doesn't close our transport when garbage collected
        global _WRITER_GRAVEYARD

        _WRITER_GRAVEYARD.append(writer)
        _WRITER_GRAVEYARD = [
            w
            for w in _WRITER_GRAVEYARD
            if w.transport is not None and not w.transport.is_closing()
        ]

        # Reset the old protocol
        transport.set_protocol(old_protocol)

        # Send our reader's buffer to the old protocol
        data = bytes(reader._buffer)

        if data:
            _LOGGER.debug("Sending remaining reader data to old protocol: %r", data)
            old_protocol.data_received(data)



import contextlib

import bellows.config
import bellows.ezsp
import bellows.types
from bellows.zigbee.application import ControllerApplication
import zigpy.config


@contextlib.asynccontextmanager
async def connect_ezsp(port: str, baudrate: int = 115200) -> bellows.ezsp.EZSP:
    """Context manager to return a connected EZSP instance for a serial port."""

    ezsp = bellows.ezsp.EZSP(
        # We use this roundabout way to construct the device schema to make sure that
        # we are compatible with future changes to the zigpy device config schema.
        ControllerApplication.SCHEMA(
            {
                zigpy.config.CONF_DEVICE: {
                    zigpy.config.CONF_DEVICE_PATH: port,
                    zigpy.config.CONF_DEVICE_BAUDRATE: baudrate,
                }
            }
        )[zigpy.config.CONF_DEVICE]
    )

    await ezsp.connect(use_thread=False)

    try:
        yield ezsp
    finally:
        await ezsp.disconnect()



import pytest

from universal_silabs_flasher.common import crc16_kermit
import universal_silabs_flasher.spinel as spinel


@pytest.mark.parametrize(
    "data, crc",
    [
        (b"", 0x0000),
        (b"foobar", 0x147B),
        (b"\xfa\x9b\x51\xb9\xf2\x53\xe3\xbd", 0x6782),
    ],
)
def test_hdlc_lite_crc(data, crc):
    assert crc16_kermit(data) == crc


@pytest.mark.parametrize(
    "encoded, decoded",
    [
        (bytes.fromhex("7e810243d3d37e"), bytes.fromhex("810243")),
        (bytes.fromhex("7e8103367d5e7d5d6af97e"), bytes.fromhex("8103367e7d")),
        (bytes.fromhex("7e810365010b287e"), bytes.fromhex("81036501")),
        (bytes.fromhex("7e8103862a01547d5e7e"), bytes.fromhex("8103862a01")),
        (
            bytes.fromhex(
                "7e8106024f50454e5448524541442f366666316163302d64697274793b204546523332"
                "3b2044656320323320323032322031383a30383a303000fa8c7e"
            ),
            bytes.fromhex(
                "8106024f50454e5448524541442f366666316163302d64697274793b2045465233323b"
                "2044656320323320323032322031383a30383a303000"
            ),
        ),
    ],
)
def test_hdlc_lite_encoding_decoding(encoded, decoded):
    assert spinel.HDLCLiteFrame(data=decoded).serialize() == encoded
    assert spinel.HDLCLiteFrame.from_bytes(encoded).data == decoded


@pytest.mark.parametrize(
    "encoded, decoded",
    [
        (
            bytes.fromhex(
                "8106024f50454e5448524541442f366666316163302d64697274793b2045465233323b"
                "2044656320323320323032322031383a30383a303000"
            ),
            spinel.SpinelFrame(
                header=spinel.SpinelHeader(
                    transaction_id=1,
                    network_link_id=0,
                    flag=0b10,
                ),
                command_id=spinel.CommandID.PROP_VALUE_IS,
                data=b"\x02OPENTHREAD/6ff1ac0-dirty; EFR32; Dec 23 2022 18:08:00\x00",
            ),
        ),
    ],
)
def test_spinel_parsing(encoded, decoded):
    assert spinel.SpinelFrame.from_bytes(encoded) == decoded
    assert decoded.serialize() == encoded



import zigpy.types as zigpy_t

FLAG = zigpy_t.uint8_t(0x14)


class FrameType(zigpy_t.enum2):
    INFORMATION = 1
    SUPERVISORY = 2
    UNNUMBERED = 3


class EndpointId(zigpy_t.enum8):
    SYSTEM = 0
    SECURITY = 1
    BLUETOOTH = 2
    RAIL_DOWNSTREAM = 3
    RAIL_UPSTREAM = 4
    ZIGBEE = 5
    ZWAVE = 6
    CONNECT = 7
    GPIO = 8
    OPENTHREAD = 9
    WISUN = 10
    WIFI = 11
    DOT15_4 = 12
    CLI = 13
    BLUETOOTH_RCP = 14
    ACP = 15


class UnnumberedFrameType(zigpy_t.enum8):
    INFORMATION = 0x00
    POLL_FINAL = 0x04
    RESET_SEQ = 0x31
    ACKNOWLEDGE = 0x0E
    UNKNOWN = 0xFF


class UnnumberedFrameCommandId(zigpy_t.enum8):
    NOOP = 0x00
    RESET = 0x01
    PROP_VALUE_GET = 0x02
    PROP_VALUE_SET = 0x03
    PROP_VALUE_IS = 0x06
    INVALID = 0xFF


class PropertyId(zigpy_t.enum32):
    LAST_STATUS = 0x0000
    PROTOCOL_VERSION = 0x0001
    CAPABILITIES = 0x0002
    SECONDARY_CPC_VERSION = 0x0003
    SECONDARY_APP_VERSION = 0x0004
    RX_CAPABILITY = 0x0020
    FC_VALIDATION_VALUE = 0x0030
    BOOTLOADER_INFO = 0x0200
    BOOTLOADER_REBOOT_MODE = 0x0202
    SECURITY_STATE = 0x0301
    CORE_DEBUG_COUNTERS = 0x0400
    UFRAME_PROCESSING = 0x0500
    ENTER_IRQ = 0x0600

    ENDPOINT_STATE_0 = 0x1000
    ENDPOINT_STATE_1 = 0x1001
    ENDPOINT_STATE_2 = 0x1002
    ENDPOINT_STATE_3 = 0x1003
    ENDPOINT_STATE_4 = 0x1004
    ENDPOINT_STATE_5 = 0x1005
    ENDPOINT_STATE_6 = 0x1006
    ENDPOINT_STATE_7 = 0x1007
    ENDPOINT_STATE_8 = 0x1008
    ENDPOINT_STATE_9 = 0x1009
    ENDPOINT_STATE_10 = 0x100A
    ENDPOINT_STATE_11 = 0x100B
    ENDPOINT_STATE_12 = 0x100C
    ENDPOINT_STATE_13 = 0x100D
    ENDPOINT_STATE_14 = 0x100E
    ENDPOINT_STATE_15 = 0x100F
    ENDPOINT_STATE_16 = 0x1010
    ENDPOINT_STATE_17 = 0x1011
    ENDPOINT_STATE_18 = 0x1012
    ENDPOINT_STATE_19 = 0x1013
    ENDPOINT_STATE_20 = 0x1014
    ENDPOINT_STATE_21 = 0x1015
    ENDPOINT_STATE_22 = 0x1016
    ENDPOINT_STATE_23 = 0x1017
    ENDPOINT_STATE_24 = 0x1018
    ENDPOINT_STATE_25 = 0x1019
    ENDPOINT_STATE_26 = 0x101A
    ENDPOINT_STATE_27 = 0x101B
    ENDPOINT_STATE_28 = 0x101C
    ENDPOINT_STATE_29 = 0x101D
    ENDPOINT_STATE_30 = 0x101E
    ENDPOINT_STATE_31 = 0x101F
    ENDPOINT_STATE_32 = 0x1020
    ENDPOINT_STATE_33 = 0x1021
    ENDPOINT_STATE_34 = 0x1022
    ENDPOINT_STATE_35 = 0x1023
    ENDPOINT_STATE_36 = 0x1024
    ENDPOINT_STATE_37 = 0x1025
    ENDPOINT_STATE_38 = 0x1026
    ENDPOINT_STATE_39 = 0x1027
    ENDPOINT_STATE_40 = 0x1028
    ENDPOINT_STATE_41 = 0x1029
    ENDPOINT_STATE_42 = 0x102A
    ENDPOINT_STATE_43 = 0x102B
    ENDPOINT_STATE_44 = 0x102C
    ENDPOINT_STATE_45 = 0x102D
    ENDPOINT_STATE_46 = 0x102E
    ENDPOINT_STATE_47 = 0x102F
    ENDPOINT_STATE_48 = 0x1030
    ENDPOINT_STATE_49 = 0x1031
    ENDPOINT_STATE_50 = 0x1032
    ENDPOINT_STATE_51 = 0x1033
    ENDPOINT_STATE_52 = 0x1034
    ENDPOINT_STATE_53 = 0x1035
    ENDPOINT_STATE_54 = 0x1036
    ENDPOINT_STATE_55 = 0x1037
    ENDPOINT_STATE_56 = 0x1038
    ENDPOINT_STATE_57 = 0x1039
    ENDPOINT_STATE_58 = 0x103A
    ENDPOINT_STATE_59 = 0x103B
    ENDPOINT_STATE_60 = 0x103C
    ENDPOINT_STATE_61 = 0x103D
    ENDPOINT_STATE_62 = 0x103E
    ENDPOINT_STATE_63 = 0x103F
    ENDPOINT_STATE_64 = 0x1040
    ENDPOINT_STATE_65 = 0x1041
    ENDPOINT_STATE_66 = 0x1042
    ENDPOINT_STATE_67 = 0x1043
    ENDPOINT_STATE_68 = 0x1044
    ENDPOINT_STATE_69 = 0x1045
    ENDPOINT_STATE_70 = 0x1046
    ENDPOINT_STATE_71 = 0x1047
    ENDPOINT_STATE_72 = 0x1048
    ENDPOINT_STATE_73 = 0x1049
    ENDPOINT_STATE_74 = 0x104A
    ENDPOINT_STATE_75 = 0x104B
    ENDPOINT_STATE_76 = 0x104C
    ENDPOINT_STATE_77 = 0x104D
    ENDPOINT_STATE_78 = 0x104E
    ENDPOINT_STATE_79 = 0x104F
    ENDPOINT_STATE_80 = 0x1050
    ENDPOINT_STATE_81 = 0x1051
    ENDPOINT_STATE_82 = 0x1052
    ENDPOINT_STATE_83 = 0x1053
    ENDPOINT_STATE_84 = 0x1054
    ENDPOINT_STATE_85 = 0x1055
    ENDPOINT_STATE_86 = 0x1056
    ENDPOINT_STATE_87 = 0x1057
    ENDPOINT_STATE_88 = 0x1058
    ENDPOINT_STATE_89 = 0x1059
    ENDPOINT_STATE_90 = 0x105A
    ENDPOINT_STATE_91 = 0x105B
    ENDPOINT_STATE_92 = 0x105C
    ENDPOINT_STATE_93 = 0x105D
    ENDPOINT_STATE_94 = 0x105E
    ENDPOINT_STATE_95 = 0x105F
    ENDPOINT_STATE_96 = 0x1060
    ENDPOINT_STATE_97 = 0x1061
    ENDPOINT_STATE_98 = 0x1062
    ENDPOINT_STATE_99 = 0x1063
    ENDPOINT_STATE_100 = 0x1064
    ENDPOINT_STATE_101 = 0x1065
    ENDPOINT_STATE_102 = 0x1066
    ENDPOINT_STATE_103 = 0x1067
    ENDPOINT_STATE_104 = 0x1068
    ENDPOINT_STATE_105 = 0x1069
    ENDPOINT_STATE_106 = 0x106A
    ENDPOINT_STATE_107 = 0x106B
    ENDPOINT_STATE_108 = 0x106C
    ENDPOINT_STATE_109 = 0x106D
    ENDPOINT_STATE_110 = 0x106E
    ENDPOINT_STATE_111 = 0x106F
    ENDPOINT_STATE_112 = 0x1070
    ENDPOINT_STATE_113 = 0x1071
    ENDPOINT_STATE_114 = 0x1072
    ENDPOINT_STATE_115 = 0x1073
    ENDPOINT_STATE_116 = 0x1074
    ENDPOINT_STATE_117 = 0x1075
    ENDPOINT_STATE_118 = 0x1076
    ENDPOINT_STATE_119 = 0x1077
    ENDPOINT_STATE_120 = 0x1078
    ENDPOINT_STATE_121 = 0x1079
    ENDPOINT_STATE_122 = 0x107A
    ENDPOINT_STATE_123 = 0x107B
    ENDPOINT_STATE_124 = 0x107C
    ENDPOINT_STATE_125 = 0x107D
    ENDPOINT_STATE_126 = 0x107E
    ENDPOINT_STATE_127 = 0x107F
    ENDPOINT_STATE_128 = 0x1080
    ENDPOINT_STATE_129 = 0x1081
    ENDPOINT_STATE_130 = 0x1082
    ENDPOINT_STATE_131 = 0x1083
    ENDPOINT_STATE_132 = 0x1084
    ENDPOINT_STATE_133 = 0x1085
    ENDPOINT_STATE_134 = 0x1086
    ENDPOINT_STATE_135 = 0x1087
    ENDPOINT_STATE_136 = 0x1088
    ENDPOINT_STATE_137 = 0x1089
    ENDPOINT_STATE_138 = 0x108A
    ENDPOINT_STATE_139 = 0x108B
    ENDPOINT_STATE_140 = 0x108C
    ENDPOINT_STATE_141 = 0x108D
    ENDPOINT_STATE_142 = 0x108E
    ENDPOINT_STATE_143 = 0x108F
    ENDPOINT_STATE_144 = 0x1090
    ENDPOINT_STATE_145 = 0x1091
    ENDPOINT_STATE_146 = 0x1092
    ENDPOINT_STATE_147 = 0x1093
    ENDPOINT_STATE_148 = 0x1094
    ENDPOINT_STATE_149 = 0x1095
    ENDPOINT_STATE_150 = 0x1096
    ENDPOINT_STATE_151 = 0x1097
    ENDPOINT_STATE_152 = 0x1098
    ENDPOINT_STATE_153 = 0x1099
    ENDPOINT_STATE_154 = 0x109A
    ENDPOINT_STATE_155 = 0x109B
    ENDPOINT_STATE_156 = 0x109C
    ENDPOINT_STATE_157 = 0x109D
    ENDPOINT_STATE_158 = 0x109E
    ENDPOINT_STATE_159 = 0x109F
    ENDPOINT_STATE_160 = 0x10A0
    ENDPOINT_STATE_161 = 0x10A1
    ENDPOINT_STATE_162 = 0x10A2
    ENDPOINT_STATE_163 = 0x10A3
    ENDPOINT_STATE_164 = 0x10A4
    ENDPOINT_STATE_165 = 0x10A5
    ENDPOINT_STATE_166 = 0x10A6
    ENDPOINT_STATE_167 = 0x10A7
    ENDPOINT_STATE_168 = 0x10A8
    ENDPOINT_STATE_169 = 0x10A9
    ENDPOINT_STATE_170 = 0x10AA
    ENDPOINT_STATE_171 = 0x10AB
    ENDPOINT_STATE_172 = 0x10AC
    ENDPOINT_STATE_173 = 0x10AD
    ENDPOINT_STATE_174 = 0x10AE
    ENDPOINT_STATE_175 = 0x10AF
    ENDPOINT_STATE_176 = 0x10B0
    ENDPOINT_STATE_177 = 0x10B1
    ENDPOINT_STATE_178 = 0x10B2
    ENDPOINT_STATE_179 = 0x10B3
    ENDPOINT_STATE_180 = 0x10B4
    ENDPOINT_STATE_181 = 0x10B5
    ENDPOINT_STATE_182 = 0x10B6
    ENDPOINT_STATE_183 = 0x10B7
    ENDPOINT_STATE_184 = 0x10B8
    ENDPOINT_STATE_185 = 0x10B9
    ENDPOINT_STATE_186 = 0x10BA
    ENDPOINT_STATE_187 = 0x10BB
    ENDPOINT_STATE_188 = 0x10BC
    ENDPOINT_STATE_189 = 0x10BD
    ENDPOINT_STATE_190 = 0x10BE
    ENDPOINT_STATE_191 = 0x10BF
    ENDPOINT_STATE_192 = 0x10C0
    ENDPOINT_STATE_193 = 0x10C1
    ENDPOINT_STATE_194 = 0x10C2
    ENDPOINT_STATE_195 = 0x10C3
    ENDPOINT_STATE_196 = 0x10C4
    ENDPOINT_STATE_197 = 0x10C5
    ENDPOINT_STATE_198 = 0x10C6
    ENDPOINT_STATE_199 = 0x10C7
    ENDPOINT_STATE_200 = 0x10C8
    ENDPOINT_STATE_201 = 0x10C9
    ENDPOINT_STATE_202 = 0x10CA
    ENDPOINT_STATE_203 = 0x10CB
    ENDPOINT_STATE_204 = 0x10CC
    ENDPOINT_STATE_205 = 0x10CD
    ENDPOINT_STATE_206 = 0x10CE
    ENDPOINT_STATE_207 = 0x10CF
    ENDPOINT_STATE_208 = 0x10D0
    ENDPOINT_STATE_209 = 0x10D1
    ENDPOINT_STATE_210 = 0x10D2
    ENDPOINT_STATE_211 = 0x10D3
    ENDPOINT_STATE_212 = 0x10D4
    ENDPOINT_STATE_213 = 0x10D5
    ENDPOINT_STATE_214 = 0x10D6
    ENDPOINT_STATE_215 = 0x10D7
    ENDPOINT_STATE_216 = 0x10D8
    ENDPOINT_STATE_217 = 0x10D9
    ENDPOINT_STATE_218 = 0x10DA
    ENDPOINT_STATE_219 = 0x10DB
    ENDPOINT_STATE_220 = 0x10DC
    ENDPOINT_STATE_221 = 0x10DD
    ENDPOINT_STATE_222 = 0x10DE
    ENDPOINT_STATE_223 = 0x10DF
    ENDPOINT_STATE_224 = 0x10E0
    ENDPOINT_STATE_225 = 0x10E1
    ENDPOINT_STATE_226 = 0x10E2
    ENDPOINT_STATE_227 = 0x10E3
    ENDPOINT_STATE_228 = 0x10E4
    ENDPOINT_STATE_229 = 0x10E5
    ENDPOINT_STATE_230 = 0x10E6
    ENDPOINT_STATE_231 = 0x10E7
    ENDPOINT_STATE_232 = 0x10E8
    ENDPOINT_STATE_233 = 0x10E9
    ENDPOINT_STATE_234 = 0x10EA
    ENDPOINT_STATE_235 = 0x10EB
    ENDPOINT_STATE_236 = 0x10EC
    ENDPOINT_STATE_237 = 0x10ED
    ENDPOINT_STATE_238 = 0x10EE
    ENDPOINT_STATE_239 = 0x10EF
    ENDPOINT_STATE_240 = 0x10F0
    ENDPOINT_STATE_241 = 0x10F1
    ENDPOINT_STATE_242 = 0x10F2
    ENDPOINT_STATE_243 = 0x10F3
    ENDPOINT_STATE_244 = 0x10F4
    ENDPOINT_STATE_245 = 0x10F5
    ENDPOINT_STATE_246 = 0x10F6
    ENDPOINT_STATE_247 = 0x10F7
    ENDPOINT_STATE_248 = 0x10F8
    ENDPOINT_STATE_249 = 0x10F9
    ENDPOINT_STATE_250 = 0x10FA
    ENDPOINT_STATE_251 = 0x10FB
    ENDPOINT_STATE_252 = 0x10FC
    ENDPOINT_STATE_253 = 0x10FD
    ENDPOINT_STATE_254 = 0x10FE
    ENDPOINT_STATE_255 = 0x10FF

    # Aggregates all endpoint state information into a single field.
    ENDPOINT_STATES = 0x1100


class RebootMode(zigpy_t.enum32):
    APPLICATION = 0
    BOOTLOADER = 1


class Status(zigpy_t.enum32):
    # Operation has completed successfully.
    OK = 0
    # Operation has failed for some undefined reason.
    FAILURE = 1
    # The given operation has not been implemented.
    UNIMPLEMENTED = 2
    # An argument to the given operation is invalid.
    INVALID_ARGUMENT = 3
    # The given operation is invalid for the current state of the device.
    INVALID_STATE = 4
    # The given command is not recognized.
    INVALID_COMMAND = 5
    # The given Spinel interface is not supported.
    INVALID_INTERFACE = 6
    # An internal runtime error has occurred.
    INTERNAL_ERROR = 7
    # A security or authentication error has occurred.
    SECURITY_ERROR = 8
    # An error has occurred while parsing the command.
    PARSE_ERROR = 9
    # The operation is in progress and will be completed asynchronously.
    IN_PROGRESS = 10
    # The operation has been prevented due to memory pressure.
    NOMEM = 11
    # The device is currently performing a mutually exclusive operation.
    BUSY = 12
    # The given property is not recognized.
    PROP_NOT_FOUND = 13
    # The packet was dropped.
    PACKET_DROPPED = 14
    # The result of the operation is empty.
    EMPTY = 15
    # The command was too large to fit in the internal buffer.
    CMD_TOO_BIG = 16
    # The packet was not acknowledged.
    NO_ACK = 17
    # The packet was not sent due to a CCA failure.
    CCA_FAILURE = 18
    # The operation is already in progress or the property was already set to the given
    # value.
    ALREADY = 19
    # The given item could not be found in the property.
    ITEM_NOT_FOUND = 20
    # The given command cannot be performed on this property.
    INVALID_COMMAND_FOR_PROP = 21
    # 22-111 : RESERVED
    RESET_POWER_ON = 112
    RESET_EXTERNAL = 113
    RESET_SOFTWARE = 114
    RESET_FAULT = 115
    RESET_CRASH = 116
    RESET_ASSERT = 117
    RESET_OTHER = 118
    RESET_UNKNOWN = 119
    RESET_WATCHDOG = 120



import asyncio

import pytest

from universal_silabs_flasher.common import StateMachine, Version, put_first


async def test_state_machine_bad_initial_state():
    with pytest.raises(ValueError):
        StateMachine(states={"a", "b"}, initial="c")


async def test_state_machine():
    sm = StateMachine(states={"a", "b", "c"}, initial="a")
    assert sm.state == "a"

    sm.state = "b"
    assert sm.state == "b"

    # Invalid states are ignored
    with pytest.raises(ValueError):
        sm.state = "x"

    assert sm.state == "b"

    asyncio.get_running_loop().call_later(0.01, setattr, sm, "state", "a")
    await asyncio.gather(sm.wait_for_state("a"), sm.wait_for_state("a"))

    assert sm.state == "a"


def test_put_first():
    assert put_first([1, 2, 3], [2]) == [2, 1, 3]
    assert put_first([1, 2, 3], [4]) == [4, 1, 2, 3]
    assert put_first([1, 2, 3], [1]) == [1, 2, 3]
    assert put_first([1, 2, 3], [3]) == [3, 1, 2]


@pytest.mark.parametrize(
    "version",
    [
        "2.00.01",
        "7.2.2.0 build 190",
        "4.2.2",
        "SL-OPENTHREAD/2.2.2.0_GitHub-91fa1f455",
    ],
)
def test_version_parsing(version):
    v1 = Version(version)
    v2 = Version(version)

    assert v1.compatible_with(v2)
    assert v1 == v2
    assert v1 >= v2
    assert not v1 > v2


def test_version_comparison_simple():
    assert Version("2.00.01") > Version("2.00.00")
    assert Version("2.10.01") > Version("2.00.02")
    assert Version("2.00.01") >= Version("2.00.00")
    assert Version("2.00.01") != Version("2.00.00")


def test_version_comparison_thread():
    assert Version("SL-OPENTHREAD/2.2.2.0_GitHub-91fa1f455").compatible_with(
        Version("SL-OPENTHREAD/2.2.2.0_GitHub-asdfoo")
    )

    assert not Version("SL-OPENTHREAD/2.2.2.1_GitHub-91fa1f455").compatible_with(
        Version("SL-OPENTHREAD/2.2.2.0_GitHub-asdfoo")
    )

    assert Version("SL-OPENTHREAD/2.2.2.1_GitHub-91fa1f455") > Version(
        "SL-OPENTHREAD/2.2.2.0_GitHub-asdfoo"
    )


def test_version_comparison_ezsp():
    assert Version("7.2.2.0 build 191") > Version("7.2.2.0 build 190")
    assert Version("7.2.2.0").compatible_with(Version("7.2.2.0 build 190"))
    assert not Version("7.2.2.0 build 191").compatible_with(
        Version("7.2.2.0 build 190")
    )



import asyncio
from unittest.mock import call, patch

import zigpy.types as t

from universal_silabs_flasher.common import Version
from universal_silabs_flasher.flasher import Flasher, ProbeResult


async def test_write_emberznet_eui64():
    flasher = Flasher(device="/dev/ttyMOCK")

    with (
        patch.object(
            flasher, "probe_gecko_bootloader", side_effect=asyncio.TimeoutError
        ),
        patch.object(
            flasher,
            "probe_ezsp",
            return_value=ProbeResult(
                version=Version("7.4.4.0 build 0"),
                continue_probing=False,
                baudrate=115200,
            ),
        ),
        patch.object(flasher, "_connect_ezsp") as mock_connect_ezsp,
    ):
        ezsp = mock_connect_ezsp.return_value.__aenter__.return_value

        ezsp.getEui64.return_value = (t.EUI64.convert("00:11:22:33:44:55:66:77"),)
        ezsp.write_custom_eui64.return_value = None

        await flasher.write_emberznet_eui64(
            new_ieee=t.EUI64.convert("11:22:33:44:55:66:77:88"), force=True
        )

    assert ezsp.write_custom_eui64.mock_calls == [
        call(ieee=t.EUI64.convert("11:22:33:44:55:66:77:88"), burn_into_userdata=True)
    ]



from __future__ import annotations

import asyncio
import collections
import contextlib
import dataclasses
import functools
import logging
import re
import sys
import typing

import click
import crc
import zigpy.serial

if sys.version_info[:2] < (3, 11):
    from async_timeout import timeout as asyncio_timeout  # pragma: no cover
else:
    from asyncio import timeout as asyncio_timeout  # pragma: no cover

if typing.TYPE_CHECKING:
    from typing_extensions import Self

_LOGGER = logging.getLogger(__name__)

CONNECT_TIMEOUT = 1
PROBE_TIMEOUT = 2


CRC_CCITT = crc.Calculator(
    crc.Configuration(
        width=16,
        polynomial=0x1021,
        init_value=0x0000,
        final_xor_value=0x0000,
        reverse_input=False,
        reverse_output=False,
    )
)

CRC_KERMIT = crc.Calculator(
    crc.Configuration(
        width=16,
        polynomial=0x1021,
        init_value=0xFFFF,
        final_xor_value=0xFFFF,
        reverse_input=True,
        reverse_output=True,
    )
)


# Used by both CPC and XModem
def crc16_ccitt(data: bytes) -> int:
    return CRC_CCITT.checksum(data)


# Used by HDLC-Lite
def crc16_kermit(data: bytes) -> int:
    return CRC_KERMIT.checksum(data)


def pad_to_multiple(data: bytes, multiple: int, padding: bytes) -> bytes:
    assert len(padding) == 1

    if len(data) % multiple == 0:
        return data

    num_complete_blocks = len(data) // multiple
    padded_size = multiple * (num_complete_blocks + 1)

    return data + padding * (padded_size - len(data))


class BufferTooShort(Exception):
    """Protocol buffer requires more data to parse a packet."""


class StateMachine:
    """Asyncio-friendly state machine."""

    def __init__(self, states: set[str], initial: str) -> None:
        if initial not in states:
            raise ValueError(f"Unknown initial state {initial!r}: expected {states!r}")

        self._states = states
        self._state = initial

        self._futures_for_state: collections.defaultdict[str, list[asyncio.Future]] = (
            collections.defaultdict(list)
        )

    @property
    def state(self) -> str:
        return self._state

    @state.setter
    def state(self, state: str) -> None:
        if state not in self._states:
            raise ValueError(f"Unknown state {state!r}: expected {self._states!r}")

        self._state = state

        for future in self._futures_for_state[state]:
            future.set_result(None)

    async def wait_for_state(self, state: str) -> None:
        """Waits for a state. Returns immediately if the state is active."""
        assert state in self._states

        if self.state == state:
            return

        future = asyncio.get_running_loop().create_future()
        self._futures_for_state[state].append(future)

        try:
            return await future
        finally:
            # Always clean up the future
            self._futures_for_state[state].remove(future)


@contextlib.asynccontextmanager
async def connect_protocol(port, baudrate, factory):
    loop = asyncio.get_running_loop()

    async with asyncio_timeout(CONNECT_TIMEOUT):
        _, protocol = await zigpy.serial.create_serial_connection(
            loop=loop,
            protocol_factory=factory,
            url=port,
            baudrate=baudrate,
        )
        await protocol.wait_until_connected()

    try:
        yield protocol
    finally:
        await protocol.disconnect()


class CommaSeparatedNumbers(click.ParamType):
    """Click type to parse comma-separated numbers into a list of integers."""

    name = "numbers"

    def convert(
        self, value: typing.Any, param: click.Parameter | None, ctx: click.Context
    ) -> list[int]:
        if isinstance(value, list):
            return value

        values = []

        for v in value.split(","):
            if not v.strip():
                continue

            try:
                values.append(int(v, 10))
            except ValueError:
                raise click.BadParameter(
                    f"Comma-separated list of numbers contains bad value: {v!r}"
                )

        return values


def put_first(lst: list[typing.Any], elements: list[typing.Any]) -> list[typing.Any]:
    """Orders a list so that the provided element is first."""
    return elements + [e for e in lst if e not in elements]


@dataclasses.dataclass(frozen=True, order=True)
class VersionComponent:
    comparable: bool
    data: str | int


@functools.total_ordering
class Version:
    _SEPARATORS = {".", "-", "/", "_", " build "}
    _SEPARATORS_REGEX = re.compile(
        "(" + "|".join(re.escape(s) for s in _SEPARATORS) + ")"
    )

    def __init__(self, version: str) -> None:
        self.components: list[VersionComponent] = []
        # 2.00.01
        # 7.2.2.0 build 190
        # 4.2.2
        # SL-OPENTHREAD/2.2.2.0_GitHub-91fa1f455
        # 4.4.0-2546d625-dirty-676fdb09
        for component in self._SEPARATORS_REGEX.split(version):
            if component.isdigit():
                self.components.append(
                    VersionComponent(comparable=True, data=int(component))
                )
            else:
                self.components.append(
                    VersionComponent(comparable=False, data=component)
                )

    def comparable_components(self) -> tuple[VersionComponent, ...]:
        return tuple(c for c in self.components if c.comparable)

    def compatible_with(self, other: Self) -> bool:
        our_comparable = self.comparable_components()
        their_comparable = other.comparable_components()

        prefix_length = min(len(our_comparable), len(their_comparable))
        return our_comparable[:prefix_length] == their_comparable[:prefix_length]

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Version):
            return NotImplemented

        return self.components == other.components

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, Version):
            return NotImplemented

        our_comparable = self.comparable_components()
        their_comparable = other.comparable_components()

        return our_comparable < their_comparable

    def __repr__(self) -> str:
        concatenated = "".join(str(c.data) for c in self.components)
        comparable = ".".join(str(c.data) for c in self.comparable_components())

        if concatenated == comparable:
            return f"{concatenated!r}"

        return f"{concatenated!r} ({comparable})"



import click
import pytest

from universal_silabs_flasher.flash import SerialPort


def test_click_serialport_validation():
    assert SerialPort().convert("/dev/null", None, None) == "/dev/null"
    assert SerialPort().convert("socket://1.2.3.4", None, None) == "socket://1.2.3.4"
    assert SerialPort().convert("COM1", None, None) == "COM1"
    assert SerialPort().convert("\\\\.\\COM123", None, None) == "\\\\.\\COM123"

    with pytest.raises(click.BadParameter) as exc_info:
        assert SerialPort().convert("COM10", None, None)

    with pytest.raises(click.BadParameter) as exc_info:
        assert SerialPort().convert("http://1.2.3.4", None, None)

    assert "invalid URL scheme" in exc_info.value.message

    with pytest.raises(click.BadParameter) as exc_info:
        assert SerialPort().convert("/dev/serial/by-id/does-not-exist", None, None)

    assert "does not exist" in exc_info.value.message



from __future__ import annotations

import asyncio
import enum
import logging
import re
import typing

from zigpy.serial import SerialProtocol

from .common import PROBE_TIMEOUT, StateMachine, Version, asyncio_timeout
from .xmodemcrc import send_xmodem128_crc

_LOGGER = logging.getLogger(__name__)


class UploadError(Exception):
    pass


class NoFirmwareError(Exception):
    pass


MENU_AFTER_UPLOAD_TIMEOUT = 0.5
RUN_APPLICATION_DELAY = 0.1

MENU_REGEX = re.compile(
    rb"\r\n(?P<type>Gecko|\w+ Serial) Bootloader v(?P<version>.*?)\r\n"
    rb"1\. upload (?:gbl|ebl)\r\n"
    rb"2\. run\r\n"
    rb"3\. ebl info\r\n"
    rb"BL > "
)

UPLOAD_STATUS_REGEX = re.compile(
    rb"\r\nSerial upload (?P<status>complete|aborted)\r\n"
    rb"(?P<message>.*?)\x00?",
    flags=re.DOTALL,
)  # fmt: skip


class State(str, enum.Enum):
    WAITING_FOR_MENU = "waiting_for_menu"
    IN_MENU = "in_menu"
    WAITING_XMODEM_READY = "waiting_xmodem_ready"
    XMODEM_READY = "xmodem_ready"
    WAITING_UPLOAD_DONE = "waiting_upload_done"
    UPLOAD_DONE = "upload_done"


class GeckoBootloaderOption(bytes, enum.Enum):
    UPLOAD_FIRMWARE = b"1"
    RUN_FIRMWARE = b"2"
    EBL_INFO = b"3"


class GeckoBootloaderProtocol(SerialProtocol):
    def __init__(self) -> None:
        super().__init__()
        self._state_machine = StateMachine(
            states=list(State),
            initial=State.WAITING_FOR_MENU,
        )
        self._version: str | None = None
        self._upload_status: str | None = None

    async def probe(self) -> Version:
        """Attempt to communicate with the bootloader."""
        async with asyncio_timeout(PROBE_TIMEOUT):
            return await self.ebl_info()

    async def ebl_info(self) -> Version:
        """Select `ebl info` in the menu and return the bootloader version."""
        self._state_machine.state = State.WAITING_FOR_MENU

        # Ember bootloader requires a newline
        self.send_data(b"\n")
        self.send_data(GeckoBootloaderOption.EBL_INFO)

        await self._state_machine.wait_for_state(State.IN_MENU)

        assert self._version is not None
        return Version(self._version)

    async def run_firmware(self) -> None:
        """Select `run` in the menu."""
        await self._state_machine.wait_for_state(State.IN_MENU)

        # If the firmware fails to launch, the menu will appear again
        self._state_machine.state = State.WAITING_FOR_MENU
        self.send_data(GeckoBootloaderOption.RUN_FIRMWARE)

        try:
            async with asyncio_timeout(RUN_APPLICATION_DELAY):
                await self._state_machine.wait_for_state(State.IN_MENU)
        except asyncio.TimeoutError:
            # The menu did not appear so the application must be running
            return
        else:
            raise NoFirmwareError("No firmware exists on the device")

    async def upload_firmware(
        self,
        firmware: bytes,
        *,
        max_failures: int = 3,
        progress_callback: typing.Callable[[int, int], typing.Any] | None = None,
    ) -> None:
        """Select `upload gbl` in the menu and upload GBL firmware."""
        await self.ebl_info()

        # Select the option
        self._state_machine.state = State.WAITING_XMODEM_READY
        self.send_data(GeckoBootloaderOption.UPLOAD_FIRMWARE)

        # Wait for the XMODEM `C` byte
        await self._state_machine.wait_for_state(State.XMODEM_READY)

        # Swap protocols and transfer the data
        self._upload_status = None
        self._state_machine.state = State.WAITING_UPLOAD_DONE

        await send_xmodem128_crc(
            firmware,
            transport=self._transport,
            max_failures=max_failures,
            progress_callback=progress_callback,
        )

        await self._state_machine.wait_for_state(State.UPLOAD_DONE)
        self._state_machine.state = State.WAITING_FOR_MENU

        # The menu is sometimes sent immediately after upload
        try:
            async with asyncio_timeout(MENU_AFTER_UPLOAD_TIMEOUT):
                await self._state_machine.wait_for_state(State.IN_MENU)
        except asyncio.TimeoutError:
            # If not, trigger it manually
            await self.ebl_info()

        if self._upload_status != "complete":
            raise UploadError(self._upload_status)

    def send_data(self, data: bytes) -> None:
        assert self._transport is not None
        _LOGGER.debug("Sending data %s", data)
        self._transport.write(data)

    def data_received(self, data: bytes) -> None:
        super().data_received(data)

        while self._buffer:
            _LOGGER.debug("Parsing %s: %r", self._state_machine.state, self._buffer)
            if self._state_machine.state == State.WAITING_FOR_MENU:
                match = MENU_REGEX.search(self._buffer)

                if match is None:
                    return

                self._version = match.group("version").decode("ascii")
                _LOGGER.debug("Detected version string %r", self._version)

                self._buffer.clear()
                self._state_machine.state = State.IN_MENU
            elif self._state_machine.state == State.WAITING_XMODEM_READY:
                if not self._buffer.endswith(b"C"):
                    break

                self._buffer.clear()
                self._state_machine.state = State.XMODEM_READY
            elif self._state_machine.state == State.WAITING_UPLOAD_DONE:
                match = UPLOAD_STATUS_REGEX.search(self._buffer)

                if match is None:
                    return

                status = match.group("status").decode("ascii")

                if status == "complete":
                    self._upload_status = status
                else:
                    self._upload_status = match.group("message").decode("ascii")

                del self._buffer[: match.span()[1]]
                self._state_machine.state = State.UPLOAD_DONE

                _LOGGER.debug("Upload status: %s", self._upload_status)
            else:
                # Ignore data otherwise
                break



from __future__ import annotations

import asyncio
import dataclasses
import logging
import typing

from zigpy.serial import SerialProtocol
import zigpy.types

from .common import Version, asyncio_timeout, crc16_kermit
from .spinel_types import CommandID, HDLCSpecial, PropertyID, ResetReason

_LOGGER = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True)
class HDLCLiteFrame:
    data: bytes

    def serialize(self) -> bytes:
        payload = self.data + crc16_kermit(self.data).to_bytes(2, "little")
        encoded = bytearray()

        for byte in payload:
            if byte in (
                HDLCSpecial.FLAG,
                HDLCSpecial.ESCAPE,
                HDLCSpecial.XON,
                HDLCSpecial.XOFF,
                HDLCSpecial.VENDOR,
            ):
                encoded.append(HDLCSpecial.ESCAPE)
                byte ^= 0x20

            encoded.append(byte)

        return bytes([HDLCSpecial.FLAG]) + bytes(encoded) + bytes([HDLCSpecial.FLAG])

    @classmethod
    def from_bytes(cls, data: bytes) -> HDLCLiteFrame:
        unescaped = bytearray()
        unescaping = False

        for byte in data:
            if unescaping:
                byte ^= 0x20

                if byte not in (
                    HDLCSpecial.FLAG,
                    HDLCSpecial.ESCAPE,
                    HDLCSpecial.XON,
                    HDLCSpecial.XOFF,
                    HDLCSpecial.VENDOR,
                ):
                    raise ValueError(f"Invalid unescaped byte: 0x{byte:02X}")

                unescaping = False
            elif byte == HDLCSpecial.ESCAPE:
                unescaping = True
                continue
            elif byte == HDLCSpecial.FLAG:
                continue

            unescaped.append(byte)

        data = unescaped[:-2]
        crc = unescaped[-2:]
        computed_crc = crc16_kermit(data).to_bytes(2, "little")

        if computed_crc != crc:
            raise ValueError(f"Invalid CRC-16: expected {crc!r}, got {computed_crc!r}")

        return cls(data=bytes(data))


class SpinelHeader(zigpy.types.Struct):
    # TODO: allow specifying struct endianness
    transaction_id: zigpy.types.uint4_t
    network_link_id: zigpy.types.uint2_t
    flag: zigpy.types.uint2_t


@dataclasses.dataclass(frozen=True)
class SpinelFrame:
    header: SpinelHeader
    command_id: CommandID
    data: bytes

    @classmethod
    def from_bytes(cls, data: bytes) -> SpinelFrame:
        orig_data = data
        header, data = SpinelHeader.deserialize(data)

        if header.flag != 0b10:
            raise ValueError(f"Spinel header flag is invalid in frame: {orig_data!r}")

        command_id, data = CommandID.deserialize(data)

        return cls(header=header, command_id=command_id, data=data)

    def serialize(self) -> bytes:
        return self.header.serialize() + self.command_id.serialize() + self.data


class SpinelProtocol(SerialProtocol):
    _buffer: bytearray

    def __init__(self) -> None:
        super().__init__()
        self._transaction_id: int = 1
        self._pending_frames: dict[int, asyncio.Future] = {}

    def send_data(self, data: bytes) -> None:
        assert self._transport is not None
        _LOGGER.debug("Sending data %s", data)
        self._transport.write(data)

    def data_received(self, data: bytes) -> None:
        super().data_received(data)

        self._buffer = self._buffer.lstrip(bytes([HDLCSpecial.FLAG]))

        if bytes([HDLCSpecial.FLAG]) not in self._buffer:
            return

        while self._buffer:
            # Flag bytes can come before and after any packet, any number of times
            chunk, _, self._buffer = self._buffer.partition(bytes([HDLCSpecial.FLAG]))

            if not chunk:
                continue

            # Decode the HDLC frame
            try:
                hdlc_frame = HDLCLiteFrame.from_bytes(chunk)
            except ValueError:
                _LOGGER.debug("Failed to decode HDLC chunk %r", chunk)
                continue

            _LOGGER.debug("Decoded HDLC frame: %r", hdlc_frame)

            # And finally the Spinel frame
            try:
                spinel_frame = SpinelFrame.from_bytes(hdlc_frame.data)
            except ValueError as e:
                _LOGGER.debug("Failed to decode Spinel frame: %r", e)
                continue

            self.frame_received(spinel_frame)

    def frame_received(self, frame: SpinelFrame) -> None:
        _LOGGER.debug("Parsed frame %r", frame)

        if frame.header.transaction_id in self._pending_frames:
            self._pending_frames[frame.header.transaction_id].set_result(frame)

    @typing.overload
    async def send_frame(
        self,
        frame: SpinelFrame,
        *,
        wait_response: typing.Literal[True],
        retries: int,
        timeout: float,
        retry_delay: float,
    ) -> None: ...

    @typing.overload
    async def send_frame(
        self,
        frame: SpinelFrame,
        *,
        wait_response: typing.Literal[False],
        retries: int,
        timeout: float,
        retry_delay: float,
    ) -> SpinelFrame: ...

    async def send_frame(
        self,
        frame: SpinelFrame,
        *,
        wait_response: bool = True,
        retries: int = 3,
        timeout: float = 1,
        retry_delay: float = 0.1,
    ) -> SpinelFrame | None:
        # A transaction ID of `0` is special: we only use 1-15
        self._transaction_id = (self._transaction_id + 1) % (0b1111 - 1)
        tid = 1 + self._transaction_id

        future = asyncio.get_running_loop().create_future()
        self._pending_frames[tid] = future

        # Replace the transaction ID
        new_frame = dataclasses.replace(
            frame, header=frame.header.replace(transaction_id=tid)
        )

        if not wait_response:
            _LOGGER.debug("Sending frame %r", new_frame)
            self.send_data(HDLCLiteFrame(data=new_frame.serialize()).serialize())
            return None

        try:
            for attempt in range(retries + 1):
                _LOGGER.debug("Sending frame %r", new_frame)
                self.send_data(HDLCLiteFrame(data=new_frame.serialize()).serialize())

                try:
                    async with asyncio_timeout(timeout):
                        return await asyncio.shield(future)
                except asyncio.TimeoutError:
                    _LOGGER.debug(
                        "Failed to send %s, trying again in %0.2fs (attempt %s of %s)",
                        frame,
                        retry_delay,
                        attempt + 1,
                        retries + 1,
                    )

                    if attempt >= retries:
                        raise

                    await asyncio.sleep(retry_delay)
        finally:
            del self._pending_frames[tid]

        raise AssertionError("Unreachable")

    async def send_command(
        self, command_id: CommandID, data: bytes, **kwargs
    ) -> SpinelFrame:
        frame = SpinelFrame(
            header=SpinelHeader(
                flag=0b10,
                network_link_id=0,
                transaction_id=None,
            ),
            command_id=command_id,
            data=data,
        )

        return await self.send_frame(frame, **kwargs)

    async def probe(self) -> Version:
        rsp = await self.send_command(
            CommandID.PROP_VALUE_GET,
            PropertyID.NCP_VERSION.serialize(),
        )

        prop_id, version_string = PropertyID.deserialize(rsp.data)
        assert prop_id == PropertyID.NCP_VERSION

        # SL-OPENTHREAD/2.2.2.0_GitHub-91fa1f455; EFR32; Mar 14 2023 16:03:40
        version = version_string.rstrip(b"\x00").decode("ascii")

        # We strip off the date code to get something reasonably stable
        short_version, _ = version.split(";", 1)

        return Version(short_version)

    async def enter_bootloader(self) -> None:
        await self.send_command(
            CommandID.RESET,
            ResetReason.BOOTLOADER.serialize(),
            wait_response=False,
        )



from __future__ import annotations

import asyncio
import dataclasses
import logging
import typing

import bellows.config
import bellows.ezsp
import bellows.types
from zigpy.serial import SerialProtocol
import zigpy.types

from .common import (
    PROBE_TIMEOUT,
    Version,
    asyncio_timeout,
    connect_protocol,
    pad_to_multiple,
)
from .const import DEFAULT_BAUDRATES, GPIO_CONFIGS, ApplicationType, ResetTarget
from .cpc import CPCProtocol
from .emberznet import connect_ezsp
from .firmware import FirmwareImage
from .gecko_bootloader import GeckoBootloaderProtocol, NoFirmwareError
from .gpio import find_gpiochip_by_label, send_gpio_pattern
from .spinel import SpinelProtocol
from .xmodemcrc import BLOCK_SIZE as XMODEM_BLOCK_SIZE

_LOGGER = logging.getLogger(__name__)

EZSP_BOOTLOADER_LAUNCH_DELAY = 5


@dataclasses.dataclass(frozen=True)
class ProbeResult:
    version: Version | None
    continue_probing: bool
    baudrate: int


class Flasher:
    def __init__(
        self,
        *,
        baudrates: dict[ApplicationType, list[int]] = DEFAULT_BAUDRATES,
        probe_methods: tuple[ApplicationType, ...] = (
            ApplicationType.GECKO_BOOTLOADER,
            ApplicationType.CPC,
            ApplicationType.EZSP,
            ApplicationType.SPINEL,
        ),
        device: str,
        bootloader_reset: str | None = None,
    ):
        self._baudrates = baudrates
        self._probe_methods = probe_methods
        self._device = device

        self.app_type: ApplicationType | None = None
        self.app_version: Version | None = None
        self.app_baudrate: int | None = None
        self.bootloader_baudrate: int | None = None

        self._reset_target: ResetTarget | None = (
            ResetTarget(bootloader_reset) if bootloader_reset else None
        )

    async def enter_bootloader_reset(self, target):
        _LOGGER.info(f"Triggering {target.value} bootloader")
        if target in GPIO_CONFIGS.keys():
            config = GPIO_CONFIGS[target]
            if "chip" not in config.keys():
                _LOGGER.warning(
                    f"When using {target.value} bootloader reset "
                    + "ensure no other CP2102 USB serial devices are connected."
                )
                config["chip"] = await find_gpiochip_by_label(config["chip_name"])
            await send_gpio_pattern(
                config["chip"], config["pin_states"], config["toggle_delay"]
            )
        else:
            await self.enter_serial_bootloader()

    async def enter_serial_bootloader(self):
        baudrate = self._baudrates[ApplicationType.GECKO_BOOTLOADER][0]
        async with connect_protocol(self._device, baudrate, SerialProtocol) as sonoff:
            serial = sonoff._transport.serial
            serial.dtr = False
            serial.rts = True
            await asyncio.sleep(0.1)
            serial.dtr = True
            serial.rts = False
            await asyncio.sleep(0.5)
            serial.dtr = False

    def _connect_gecko_bootloader(self, baudrate: int):
        return connect_protocol(self._device, baudrate, GeckoBootloaderProtocol)

    def _connect_cpc(self, baudrate: int):
        return connect_protocol(self._device, baudrate, CPCProtocol)

    def _connect_ezsp(self, baudrate: int):
        return connect_ezsp(self._device, baudrate)

    def _connect_spinel(self, baudrate: int):
        return connect_protocol(self._device, baudrate, SpinelProtocol)

    async def probe_gecko_bootloader(
        self, *, baudrate: int, run_firmware: bool = True
    ) -> ProbeResult:
        try:
            async with self._connect_gecko_bootloader(baudrate) as gecko:
                bootloader_version = await gecko.probe()

                if run_firmware:
                    await gecko.run_firmware()
                    _LOGGER.info("Launched application from bootloader")
        except NoFirmwareError:
            _LOGGER.warning("No application can be launched")
            return ProbeResult(
                version=bootloader_version,
                baudrate=baudrate,
                continue_probing=False,
            )
        else:
            return ProbeResult(
                version=bootloader_version,
                baudrate=baudrate,
                continue_probing=run_firmware,
            )

    async def probe_cpc(self, baudrate: int) -> ProbeResult:
        async with self._connect_cpc(baudrate) as cpc:
            version = await cpc.probe()

        return ProbeResult(
            version=version,
            baudrate=baudrate,
            continue_probing=False,
        )

    async def probe_ezsp(self, baudrate: int) -> ProbeResult:
        async with self._connect_ezsp(baudrate) as ezsp:
            _, _, version = await ezsp.get_board_info()

        return ProbeResult(
            version=Version(version),
            baudrate=baudrate,
            continue_probing=False,
        )

    async def probe_spinel(self, baudrate: int) -> ProbeResult:
        async with self._connect_spinel(baudrate) as spinel:
            version = await spinel.probe()

        return ProbeResult(
            version=version,
            baudrate=baudrate,
            continue_probing=False,
        )

    async def probe_app_type(
        self,
        types: typing.Iterable[ApplicationType] | None = None,
        try_first: tuple[ApplicationType, ...] = (),
    ) -> None:
        if types is None:
            types = self._probe_methods

        # fmt: off
        types = (
              [m for m in types if m in try_first]
            + [m for m in types if m not in try_first]
        )
        # fmt: on

        # Reset into bootloader
        if self._reset_target:
            await self.enter_bootloader_reset(self._reset_target)

        bootloader_probe = None

        # Only run firmware from the bootloader if we have bootloader reset and
        # other probe methods
        only_probe_bootloader = types == [ApplicationType.GECKO_BOOTLOADER]
        run_firmware = self._reset_target and not only_probe_bootloader
        probe_funcs = {
            ApplicationType.GECKO_BOOTLOADER: (
                lambda baudrate: self.probe_gecko_bootloader(
                    run_firmware=run_firmware, baudrate=baudrate
                )
            ),
            ApplicationType.CPC: self.probe_cpc,
            ApplicationType.EZSP: self.probe_ezsp,
            ApplicationType.SPINEL: self.probe_spinel,
        }

        for probe_method, baudrate in (
            (m, b) for m in types for b in self._baudrates[m]
        ):
            # Don't probe the bootloader twice
            if (
                probe_method == ApplicationType.GECKO_BOOTLOADER
                and bootloader_probe is not None
            ):
                _LOGGER.debug("Not probing bootloader twice")
                continue

            _LOGGER.info("Probing %s at %d baud", probe_method, baudrate)

            try:
                result = await probe_funcs[probe_method](baudrate=baudrate)
            except asyncio.TimeoutError:
                continue

            _LOGGER.debug("Probe result: %s", result)

            # Keep track of the bootloader version for later
            if probe_method == ApplicationType.GECKO_BOOTLOADER:
                _LOGGER.info("Detected bootloader version %s", result.version)
                bootloader_probe = result
                self.bootloader_baudrate = bootloader_probe.baudrate

            if result.continue_probing:
                continue

            self.app_type = probe_method
            self.app_version = result.version
            self.app_baudrate = result.baudrate
            break
        else:
            if bootloader_probe and self._reset_target:
                # We have no valid application image but can still re-enter the
                # bootloader
                await self.enter_bootloader_reset(self._reset_target)

                self.app_type = ApplicationType.GECKO_BOOTLOADER
                self.app_version = bootloader_probe.version
                self.app_baudrate = bootloader_probe.baudrate
                self.bootloader_baudrate = bootloader_probe.baudrate
                _LOGGER.warning("Bootloader did not launch a valid application")
            else:
                raise RuntimeError("Failed to probe running application type")

        _LOGGER.info(
            "Detected %s, version %s at %s baudrate (bootloader baudrate %s)",
            self.app_type,
            self.app_version,
            self.app_baudrate,
            self.bootloader_baudrate,
        )

    async def enter_bootloader(self) -> None:
        if self.app_type is None:
            await self.probe_app_type()

        if self.app_type is ApplicationType.GECKO_BOOTLOADER:
            # No firmware
            pass
        elif self.app_type is ApplicationType.CPC:
            async with self._connect_cpc(self.app_baudrate) as cpc:
                async with asyncio_timeout(PROBE_TIMEOUT):
                    await cpc.enter_bootloader()
        elif self.app_type is ApplicationType.SPINEL:
            async with self._connect_spinel(self.app_baudrate) as spinel:
                async with asyncio_timeout(PROBE_TIMEOUT):
                    await spinel.enter_bootloader()
        elif self.app_type is ApplicationType.EZSP:
            async with self._connect_ezsp(self.app_baudrate) as ezsp:
                try:
                    res = await ezsp.launchStandaloneBootloader(mode=0x01)
                except asyncio.TimeoutError:
                    _LOGGER.warning(
                        "Application failed to respond to bootloader launching command."
                        " Assuming bootloader has launched."
                    )
                else:
                    if res[0] != bellows.types.EmberStatus.SUCCESS:
                        raise RuntimeError(
                            f"EmberZNet could not enter the bootloader: {res[0]!r}"
                        )

                    await asyncio.sleep(EZSP_BOOTLOADER_LAUNCH_DELAY)
        else:
            raise RuntimeError(f"Invalid application type: {self.app_type}")

        # Probe the bootloader baudrate
        if self.bootloader_baudrate is None:
            await self.probe_app_type(types=[ApplicationType.GECKO_BOOTLOADER])

    async def flash_firmware(
        self,
        firmware: FirmwareImage,
        run_firmware: bool = True,
        progress_callback: typing.Callable[[int, int], typing.Any] | None = None,
    ) -> None:
        data = firmware.serialize()

        # Pad the image to the XMODEM block size
        data = pad_to_multiple(data, XMODEM_BLOCK_SIZE, b"\xff")

        async with self._connect_gecko_bootloader(self.bootloader_baudrate) as gecko:
            await gecko.probe()
            await gecko.upload_firmware(data, progress_callback=progress_callback)

            if run_firmware:
                await gecko.run_firmware()

    async def dump_emberznet_config(self) -> None:
        if self.app_type != ApplicationType.EZSP:
            raise RuntimeError(f"Device is not running EmberZNet: {self.app_type}")

        async with self._connect_ezsp(self.app_baudrate) as ezsp:
            for config in bellows.types.EzspConfigId:
                v = await ezsp.getConfigurationValue(configId=config)
                if v[0] == bellows.types.EzspStatus.ERROR_INVALID_ID:
                    continue
                print(f"{config.name}={v[1]}")

    async def write_emberznet_eui64(
        self, new_ieee: zigpy.types.EUI64, force: bool = False
    ) -> bool:
        await self.probe_app_type(
            try_first=[ApplicationType.GECKO_BOOTLOADER, ApplicationType.EZSP]
        )

        if self.app_type != ApplicationType.EZSP:
            raise RuntimeError(f"Device is not running EmberZNet: {self.app_type}")

        async with self._connect_ezsp(self.app_baudrate) as ezsp:
            (current_ieee,) = await ezsp.getEui64()
            _LOGGER.info("Current device IEEE: %s", current_ieee)

            if current_ieee == new_ieee:
                _LOGGER.info("Device IEEE address already matches, not overwriting")
                return False

            await ezsp.write_custom_eui64(ieee=new_ieee, burn_into_userdata=force)
            _LOGGER.info("Wrote new device IEEE: %s", new_ieee)

        return True



from __future__ import annotations

import asyncio
import enum
import functools
import json
import logging
import os.path
import pathlib
import re
import typing
import urllib.parse

import click
import coloredlogs
import zigpy.ota.validators
import zigpy.types

from .common import CommaSeparatedNumbers, put_first
from .const import (
    DEFAULT_BAUDRATES,
    FW_IMAGE_TYPE_TO_APPLICATION_TYPE,
    ApplicationType,
    ResetTarget,
)
from .firmware import FirmwareImageType, parse_firmware_image
from .flasher import Flasher
from .xmodemcrc import BLOCK_SIZE as XMODEM_BLOCK_SIZE, ReceiverCancelled

_LOGGER = logging.getLogger(__name__)
LOG_LEVELS = ["INFO", "DEBUG"]


def click_coroutine(f: typing.Callable) -> typing.Callable:
    @functools.wraps(f)
    def inner(*args: tuple[typing.Any], **kwargs: typing.Any) -> typing.Any:
        return asyncio.run(f(*args, **kwargs))

    return inner


def click_enum_validator_factory(
    enum_cls: type[enum.Enum],
) -> typing.Callable[[click.Context, typing.Any, typing.Any], typing.Any]:
    """Click enum validator factory."""

    def validator_callback(
        ctx: click.Context, param: click.Parameter, value: tuple[str]
    ) -> typing.Any:
        values = []

        for v in value:
            try:
                values.append(enum_cls(v))
            except ValueError:
                expected = [m.value for m in enum_cls]
                raise click.BadParameter(
                    f"{v!r} is invalid, must be one of: {', '.join(expected)}"
                )

        return values

    return validator_callback


class SerialPort(click.ParamType):
    """Click validator that accepts serial ports."""

    name = "path_or_url"

    def convert(self, value: tuple | str, param: click.Parameter, ctx: click.Context):
        if isinstance(value, tuple):
            return value

        # File
        path = pathlib.Path(value)

        if path.exists():
            return value

        # Windows COM port (COM10+ uses a different syntax)
        if re.match(r"^COM[0-9]$|\\\\\.\\COM[0-9]+$", str(path)):
            return value

        # Socket URI
        try:
            parsed = urllib.parse.urlparse(value)
        except ValueError:
            self.fail(f"Invalid URI: {path}", param, ctx)

        if parsed.scheme == "socket":
            return value
        elif parsed.scheme != "":
            self.fail(
                f"invalid URL scheme {parsed.scheme!r}, only `socket://` is accepted",
                param,
                ctx,
            )
        else:
            # Fallback
            self.fail(f"{path} does not exist", param, ctx)


@click.group()
@click.option("-v", "--verbose", count=True)
@click.option("--device", type=SerialPort())
@click.option("--baudrate", hidden=True)
@click.option(
    "--bootloader-baudrate",
    default=DEFAULT_BAUDRATES[ApplicationType.GECKO_BOOTLOADER],
    type=CommaSeparatedNumbers(),
    show_default=True,
)
@click.option(
    "--cpc-baudrate",
    default=DEFAULT_BAUDRATES[ApplicationType.CPC],
    type=CommaSeparatedNumbers(),
    show_default=True,
)
@click.option(
    "--ezsp-baudrate",
    default=DEFAULT_BAUDRATES[ApplicationType.EZSP],
    type=CommaSeparatedNumbers(),
    show_default=True,
)
@click.option(
    "--spinel-baudrate",
    default=DEFAULT_BAUDRATES[ApplicationType.SPINEL],
    type=CommaSeparatedNumbers(),
    show_default=True,
)
@click.option(
    "--probe-method",
    multiple=True,
    default=[m.value for m in ApplicationType],
    callback=click_enum_validator_factory(ApplicationType),
    show_default=True,
)
@click.option(
    "--bootloader-reset",
    type=click.Choice([t.value for t in ResetTarget]),
)
@click.pass_context
def main(
    ctx: click.Context,
    verbose: bool,
    device: str,
    baudrate: int | None,
    bootloader_baudrate: list[int],
    cpc_baudrate: list[int],
    ezsp_baudrate: list[int],
    spinel_baudrate: list[int],
    probe_method: list[ApplicationType],
    bootloader_reset: str | None,
) -> None:
    coloredlogs.install(
        fmt=(
            "%(asctime)s.%(msecs)03d"
            " %(hostname)s"
            " %(name)s"
            " %(levelname)s %(message)s"
        ),
        level=LOG_LEVELS[min(len(LOG_LEVELS) - 1, verbose)],
    )

    # Override all application baudrates if a specific value is provided
    if ctx.get_parameter_source("baudrate") != click.core.ParameterSource.DEFAULT:
        raise click.ClickException(
            "The `--baudrate` flag is deprecated. Remove it to rely on auto baudrate"
            " probing, or replace it with an application-specific baudrate flag"
            " (see `--help`)"
        )

    # To maintain some backwards compatibility, make `--device` required only when we
    # are actually invoking a command that interacts with a device
    if ctx.get_parameter_source(
        "device"
    ) == click.core.ParameterSource.DEFAULT and ctx.invoked_subcommand not in (
        dump_gbl_metadata.name
    ):
        # Replicate the "Error: Missing option" traceback
        param = next(p for p in ctx.command.params if p.name == "device")
        raise click.MissingParameter(ctx=ctx, param=param)

    ctx.obj = {
        "verbosity": verbose,
        "flasher": Flasher(
            device=device,
            baudrates={
                ApplicationType.GECKO_BOOTLOADER: bootloader_baudrate,
                ApplicationType.CPC: cpc_baudrate,
                ApplicationType.EZSP: ezsp_baudrate,
                ApplicationType.SPINEL: spinel_baudrate,
            },
            probe_methods=probe_method,
            bootloader_reset=bootloader_reset,
        ),
    }


@main.command()
@click.pass_context
@click.option("--firmware", type=click.File("rb"), required=True, show_default=True)
@click_coroutine
async def dump_gbl_metadata(ctx: click.Context, firmware: typing.BinaryIO) -> None:
    # Parse and validate the firmware image
    firmware_data = firmware.read()
    firmware.close()

    try:
        fw_image = parse_firmware_image(firmware_data)
    except zigpy.ota.validators.ValidationError as e:
        raise click.ClickException(
            f"{firmware.name!r} does not appear to be a valid firmware image: {e!r}"
        )

    try:
        metadata = fw_image.get_nabucasa_metadata()
    except KeyError:
        metadata_obj = None
    else:
        metadata_obj = metadata.original_json
        _LOGGER.info("Extracted firmware metadata: %s", metadata)

    print(json.dumps(metadata_obj))


@main.command()
@click.pass_context
@click_coroutine
async def probe(ctx: click.Context) -> None:
    flasher = ctx.obj["flasher"]

    try:
        await flasher.probe_app_type()
    except RuntimeError as e:
        raise click.ClickException(str(e)) from e

    if flasher.app_type == ApplicationType.EZSP:
        _LOGGER.info("Dumping EmberZNet Config")
        try:
            await flasher.dump_emberznet_config()
        except RuntimeError as e:
            raise click.ClickException(str(e)) from e


@main.command()
@click.pass_context
@click.option("--ieee", required=True, type=zigpy.types.EUI64.convert)
@click.option("--force", default=False, type=bool)
@click_coroutine
async def write_ieee(ctx: click.Context, ieee: zigpy.types.EUI64, force: bool) -> None:
    try:
        await ctx.obj["flasher"].write_emberznet_eui64(ieee, force=force)
    except (ValueError, RuntimeError) as e:
        raise click.ClickException(str(e)) from e


@main.command()
@click.option("--firmware", type=click.File("rb"), required=True, show_default=True)
@click.option("--force", is_flag=True, default=False, show_default=True)
@click.option("--ensure-exact-version", is_flag=True, default=False, show_default=True)
@click.option("--allow-downgrades", is_flag=True, default=False, show_default=True)
@click.option("--allow-cross-flashing", is_flag=True, default=False, show_default=True)
@click.option("--yellow-gpio-reset", is_flag=True, default=False, show_default=True)
@click.option("--sonoff-reset", is_flag=True, default=False, show_default=True)
@click.pass_context
@click_coroutine
async def flash(
    ctx: click.Context,
    firmware: typing.BinaryIO,
    force: bool,
    ensure_exact_version: bool,
    allow_downgrades: bool,
    allow_cross_flashing: bool,
    yellow_gpio_reset: bool,
    sonoff_reset: bool,
) -> None:
    flasher = ctx.obj["flasher"]

    # Parse and validate the firmware image
    firmware_data = firmware.read()
    firmware.close()

    try:
        fw_image = parse_firmware_image(firmware_data)
    except (zigpy.ota.validators.ValidationError, ValueError) as e:
        raise click.ClickException(
            f"{firmware.name!r} does not appear to be a valid firmware image: {e!r}"
        )

    try:
        metadata = fw_image.get_nabucasa_metadata()
    except Exception:
        _LOGGER.info("Failed to read firmware metadata: {exc!r}")
        metadata = None
    else:
        _LOGGER.info("Extracted GBL metadata: %s", metadata)

    # Prefer to probe with the current firmware's settings to speed up startup after the
    # firmware is flashed for the first time
    if metadata is not None and metadata.fw_type is not None:
        app_type = FW_IMAGE_TYPE_TO_APPLICATION_TYPE[metadata.fw_type]

        # Probe with the firmware's app type first
        if (
            ctx.parent.get_parameter_source("probe_method")
            == click.core.ParameterSource.DEFAULT
        ):
            _LOGGER.debug("Probing app type %s first", app_type)
            flasher._probe_methods = put_first(
                flasher._probe_methods, [ApplicationType.GECKO_BOOTLOADER, app_type]
            )

        # Probe with the firmware's baudrate first
        if (
            metadata.baudrate is not None
            and ctx.parent.get_parameter_source(f"{app_type.value}_baudrate")
            == click.core.ParameterSource.DEFAULT
        ):
            _LOGGER.debug("Probing with %s baudrate first", metadata.baudrate)
            flasher._baudrates[app_type] = put_first(
                flasher._baudrates[app_type], [metadata.baudrate]
            )
    # Maintain backward compatibility with the deprecated reset flags
    reset_msg = (
        "The '%s' flag is deprecated. Use '--bootloader-reset' "
        "instead, see --help for details."
    )
    if yellow_gpio_reset:
        flasher._reset_target = ResetTarget.YELLOW
        _LOGGER.info(reset_msg, "--yellow-gpio-reset")
    elif sonoff_reset:
        flasher._reset_target = ResetTarget.SONOFF
        _LOGGER.info(reset_msg, "--sonoff-reset")

    try:
        await flasher.probe_app_type()
    except RuntimeError as e:
        raise click.ClickException(str(e)) from e

    if flasher.app_type == ApplicationType.EZSP:
        running_image_type = FirmwareImageType.ZIGBEE_NCP
    elif flasher.app_type == ApplicationType.SPINEL:
        running_image_type = FirmwareImageType.OPENTHREAD_RCP
    elif flasher.app_type == ApplicationType.CPC:
        # TODO: how do you distinguish RCP_UART_802154 from ZIGBEE_NCP_RCP_UART_802154?
        running_image_type = FirmwareImageType.MULTIPAN
    elif flasher.app_type == ApplicationType.GECKO_BOOTLOADER:
        running_image_type = None
    else:
        raise RuntimeError(f"Unknown application type {flasher.app_type!r}")

    # Ensure the firmware versions and image types are consistent
    if not force and flasher.app_version is not None and metadata is not None:
        app_version = flasher.app_version
        fw_version = metadata.get_public_version()

        is_cross_flashing = (
            metadata.fw_type is not None
            and running_image_type is not None
            and metadata.fw_type != running_image_type
        )

        if is_cross_flashing and not allow_cross_flashing:
            raise click.ClickException(
                f"Running image type {running_image_type}"
                f" does not match firmware image type {metadata.fw_type}."
                f" If you intend to cross-flash, run with `--allow-cross-flashing`."
            )

        if not is_cross_flashing:
            if (
                metadata.baudrate is not None
                and metadata.baudrate != flasher.app_baudrate
            ):
                _LOGGER.info(
                    "Firmware baudrate %s differs from expected baudrate %s",
                    flasher.app_baudrate,
                    metadata.baudrate,
                )
            elif ensure_exact_version and app_version != fw_version:
                _LOGGER.info(
                    "Firmware version %s does not match expected version %s",
                    fw_version,
                    app_version,
                )
            elif app_version.compatible_with(fw_version):
                _LOGGER.info(
                    "Firmware version %s is flashed, not re-installing", app_version
                )
                return
            elif not allow_downgrades and app_version > fw_version:
                _LOGGER.info(
                    "Firmware version %s does not upgrade current version %s",
                    fw_version,
                    app_version,
                )
                return
        else:
            _LOGGER.info(
                "Cross-flashing from %s to %s", running_image_type, metadata.fw_type
            )

    await flasher.enter_bootloader()

    pbar = click.progressbar(
        label=os.path.basename(firmware.name),
        length=len(firmware_data),
        show_eta=True,
        show_percent=True,
    )

    # Only show the progress bar if verbose logging won't interfere
    if ctx.obj["verbosity"] > 1:
        pbar.is_hidden = True

    with pbar:
        try:
            await flasher.flash_firmware(
                fw_image,
                run_firmware=True,
                progress_callback=lambda current, _: pbar.update(XMODEM_BLOCK_SIZE),
            )
        except ReceiverCancelled:
            raise click.ClickException(
                "Firmware image was rejected by the device. Ensure this is the correct"
                " image for this device."
            )
